/* for strndup */
#define _GNU_SOURCE 1

#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <stdarg.h>
#include <fcntl.h>
#include <dkim.h>
#include <syslog.h>
#include <pwd.h>
#include <grp.h>
#include <libgen.h>
#include <time.h>


#define MAX_STORED_MSG_SIZE 524288

void log_memerr()
{
    const char * const msg = "out of memory";
    write(2, msg, strlen(msg));
    write(2, "\n", 1);
}

void logmsg(int priority, const char * const fmt, ...)
{
    va_list ap;
    va_start(ap, fmt);
    vfprintf(stderr, fmt, ap);
    fprintf(stderr, "\n");
    va_end(ap);
}

char *joinstr(const char **strs, int n_strs) {
	int i;
	int len=0,pos=0;
	char *joined;

	for(i=0;i<n_strs;i++) {
		len+=strlen(strs[i]);
	}

	/* Allocate space for strings + separator chars + terminating zero */
	joined = (char *) malloc (sizeof(char) * (len + n_strs));
	if (!joined) return NULL;
	
	for(i=0;i<n_strs;i++) {
		strcpy(joined + pos,strs[i]);
		pos+=strlen(strs[i])+1;
		if (i == n_strs - 1) break;
		joined[pos-1]=',';
	}
	
	return joined;
}

void logmsg2(int priority, const char *msg, const char *envfrom, const char **toaddrs,
		const int n_toaddrs, const char *dkimdomain,
		const char *msgid) {

	char *toaddrs_list;

	toaddrs_list = joinstr(toaddrs,n_toaddrs);
	if (!toaddrs_list) {
		log_memerr();
		return;
	}

	logmsg(priority, "%s, on mail from <%s> to <%s> dkimdomain <%s> msgid %s", msg, envfrom, toaddrs_list, dkimdomain, msgid);

	free((void *) toaddrs_list);
	
}

char *find_domain(const char *addr) {
	char *s = strchrnul(addr,'@');

	if (!s) return 0;

	if (*s == '\0') {
		return s;	
	}
	
	return s+1;
}

char * message;
int messagelen;
int message_allocated;

int canonicalise_message()
{
	char *s,*t,*r,*new_message,*p;	
	int len;
	unsigned int allocated = 0, n = 0;	

	s = message;
	len = messagelen;
	
	allocated = sizeof(char) * (len + 4096);
	r = (char *) calloc (1,allocated);
	if (!r) return 0;

	new_message = r;

	for(t=s; (t-s) < len; ) {
		if (t != s && *t == '\n') {
			if (*(t-1) != '\r') {
				*r++ = '\r';
				n++;
				if (n >= allocated) {
					allocated = sizeof(char) * (n+4096);
					p = (char *) realloc(new_message,allocated);
					if (!p) {
						free(new_message);
						return 0;
					}
					new_message = p;
					r = new_message + n;
				}
			}
		}
		*r++ = *t++;
		n++;
		if (n >= allocated) {
			allocated = sizeof(char) * (n+4096);
			p = (char *) realloc(new_message,allocated);
			if (!p) {
				free(new_message);
				return 0;
			}
			new_message = p;			
			r = new_message + n;
		}
	}

	free(message);
	message = new_message;
	messagelen = n;
	message_allocated = allocated;

	return 1;
}

int dkim_verify(const char * const message, const int messagelen, 
	unsigned char **dkim_results, char ***dkim_domains)
{
    DKIMVerifyOptions dkim_vopts;
    DKIMContext dkim_ctx;
    DKIMVerifyDetails *pDetails;
    char szPolicy[512];
    int nsigs = 0;
    int i;
    unsigned failed = 0;
    
    *dkim_results = NULL;
    *dkim_domains = NULL;

    memset(&dkim_vopts, 0, sizeof(dkim_vopts));
    dkim_vopts.pfnSelectorCallback = NULL;
    dkim_vopts.pfnPracticesCallback = NULL;

    if(DKIMVerifyInit(&dkim_ctx, &dkim_vopts) != DKIM_SUCCESS)
	return -1;

    if(DKIMVerifyProcess(&dkim_ctx, (char *)message, messagelen) == DKIM_INVALID_CONTEXT
      || DKIMVerifyResults(&dkim_ctx) == DKIM_INVALID_CONTEXT
      || DKIMVerifyGetDetails(&dkim_ctx, &nsigs, &pDetails, szPolicy)
         == DKIM_INVALID_CONTEXT)
    {
	DKIMVerifyFree(&dkim_ctx);
	return -1;
    }

    if(!pDetails)
    {
	DKIMVerifyFree(&dkim_ctx);
	return -1;
    }

    /* XXX logging */
    if (nsigs > 0) {
    	*dkim_results = (unsigned char *) malloc (sizeof(unsigned char) * nsigs);
	if (!*dkim_results) {
    		DKIMVerifyFree(&dkim_ctx);
		return -1;
	}

    	*dkim_domains = (char **) malloc (sizeof(char *) * nsigs);

	if (!*dkim_domains) {
		free(*dkim_results);
    		DKIMVerifyFree(&dkim_ctx);
		return -1;
	}

    	for (i = 0; i < nsigs; i++) {
		if (pDetails[i].nResult == DKIM_SELECTOR_DNS_TEMP_FAILURE ) {
		    printf("Result %d is temp failure\n", i);
		} else if (pDetails[i].nResult < 0) {
		    printf("Result %d is failure\n", i);
		    failed = 1;
		} else {
		    printf("Result %d is pass for domain %s\n", i, pDetails[i].IdentityDomain);
		}
    	}
    }

    DKIMVerifyFree(&dkim_ctx);
    return nsigs;

}

int main(int argc, char **argv)
{
    unsigned int header_index;

    message_allocated = 1 << 24;
    messagelen = 0;
    message = (char *)malloc(message_allocated);
    if(!message)
	log_memerr();

    FILE *fp;

    if(argc == 1)
      fp = fdopen(0, "r");
    else
      fp = fopen(argv[1], "r");

    if(!fp)
    {
      fprintf(stderr, "Can't open file\n");
      return 1;
    }

    int rc;
    while((rc = fread(&message[messagelen], 1, message_allocated - messagelen, fp)) != 0)
    {
      messagelen += rc;
    }
    fclose(fp);
    message[messagelen] = '\0';

    int dkim_signature = 0;
    if(strstr(message, "DKIM-Signature"))
        dkim_signature = 1;

    /* If the email has a DKIM signature, then verify it */
    if (dkim_signature)
    {
    	if (! canonicalise_message() ) {
		log_memerr();
		return 1;
	}
    }

    unsigned char *dkim_results;
    char **dkim_domains;
    unsigned char *dk_results;
    char **dk_domains;

    int n_dkim_sigs = 0, n_dk_sigs = 0;

    if (dkim_signature)
       n_dkim_sigs = dkim_verify(message, messagelen, &dkim_results, &dkim_domains);


}

